import os
import logging

from distutils.version import StrictVersion

import functools
import pkg_resources

from .api import event, task_event
from . import exceptions

logger = logging.getLogger(__name__)


def cert() -> str:
    """ Get the TLS certificate used to authenticate to the PPA Appliance API Proxy. The
    certificate should be used for all requests proxied through the `appliance.api_proxy`
    URL.

    Example:

        .. code-block:: python

            # sending a request to a configured Vault API
            response = requests.get(
                request_url,
                proxies={"https": appliance.api_proxy(), "http": appliance.api_proxy()},
                verify=appliance.cert(),
            )
    """
    dynamic_cert_path = "/.ppa/certs/gateway.crt"
    cert_path = (
        dynamic_cert_path
        if os.path.isfile(dynamic_cert_path)
        else pkg_resources.resource_filename("ppa", "certs/proxy.crt")
    )
    logger.debug(f"using certificate {cert_path}")
    return cert_path


def api_proxy() -> str:
    """ Get the URL of the PPA Appliance API Proxy. The proxy should be used when
    sending requests to a configured Vault API.

    Example:

        .. code-block:: python

            # sending a request to a configured Vault API
            response = requests.get(
                request_url,
                proxies={"https": appliance.api_proxy(), "http": appliance.api_proxy()},
                verify=ppa.cert(),
            )

    """
    return os.environ["PPA_API_PROXY"]


@task_event()
def apis(reason: str = "Accessing metadata of all configured APIs") -> dict:
    """ Get a list of all configured Vault APIs.

    :returns: A dictionary of Vault API configuration data. The list of API data will be
        returned in the `data` key.

    Example:

        .. code-block:: python

            response = appliance.apis(reason="Getting all configured PxM Appliances")
            api_data = response.get('data')
            pxm_appliances = [
                api for api in api_data if api.get("driver") == "osirium-pxm-rest"
            ]

    """
    return event("data/apis", reason)


def address() -> str:
    """
    :returns: The IP address of the PPA appliance the task was triggered by.
    """
    return os.environ["PPA_ADDR"]


def version() -> str:
    """
    :returns: The version of the PPA appliance the task was triggered by.
    """
    return os.environ["PPA_VERSION"]


def current_version_less_than(ver: str):
    return StrictVersion(version()) < StrictVersion(ver)


def retired(appliance_version: str, message: str = None):
    def _function_wrapper(func):
        @functools.wraps(func)
        def _check_not_retired(*args, **kwargs):
            if StrictVersion(appliance_version) <= StrictVersion(version()):
                raise exceptions.RetiredFunction(
                    f"Function {func.__name__} was retired in PPA version {appliance_version}. "
                    f"{message or 'Please view the appliance release notes for more information.'}"
                )
            return func(*args, **kwargs)

        return _check_not_retired

    return _function_wrapper


def min_ppa_version(minimum_version: str):
    def _function_wrapper(func):
        @functools.wraps(func)
        def _check_version(*args, **kwargs):
            # Minimum required version must be less than or equal to the current appliance version
            if StrictVersion(minimum_version) > StrictVersion(version()):
                raise exceptions.PPAVersionError(
                    f"Function {func.__name__} requires PPA appliance version {minimum_version} "
                    f"but the appliance version is {version()}."
                )
            return func(*args, **kwargs)

        return _check_version

    return _function_wrapper
