"""
Vaults

This module contains actions for viewing & selecting PPA Vaults in a task.
"""

import uuid
import enum
import contextlib

from typing import List, Union, Optional, Callable

from ppa import appliance, ui

from .records import Vault, VaultData
from ._common import min_ppa_version


class Vaults(enum.Enum):
    pam = "Osirium PAM Vault"
    hashicorp = "Hashicorp Vault"
    aws = "AWS Secrets Manager"
    azure = "Azure Key Vault"


class Drivers(enum.Enum):
    pam = "osirium-pxm-rest"
    hashicorp = "hashicorp-vault"
    aws = "aws-secrets-manager"
    azure = "azure-key-vault"


TABLE_HEADER = ["Name", "Type", "Address", "Expires At"]


def _generate_table_row(vault):
    return [
        vault["name"],
        getattr(Vaults, Drivers(vault["driver"]).name).value,
        vault["url"],
        vault["expires_at"],
    ]


def _raise(message: str, *, exception):
    ui.output_error(message)
    raise exception(message)


@contextlib.contextmanager
def _raise_on_empty(message):
    try:
        yield
    except StopIteration:
        _raise(message, exception=KeyError)


def get_all(reason: str = "Getting all Vaults") -> List[Vault]:
    """Get all Vaults configured in PPA.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __reason:__ text to display in the task interface (defaults to `Getting all Vaults`)

    ??? success "Output"

        A list of [Vaults](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.get_all:
          save: vaults
        ```"""
    return [VaultData(**v).__dict__ for v in appliance.apis(reason)["data"]]


def _get_by_driver(reason: str, driver: str) -> List[Vault]:
    return list(filter(lambda v: v["driver"] == driver, get_all(reason)))


def _get_by_name(name: str, *, vault_type: str, getter: Callable) -> Vault:
    with _raise_on_empty(f"No {vault_type} configured in PPA with name '{name}'."):
        return next(iter(list(filter(lambda v: v["name"] == name, getter()))))


def hashicorp_vaults(reason: str = f"Getting {Vaults.hashicorp.value}s") -> List[Vault]:
    """Get all Hashicorp Vaults configured in PPA.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __reason:__ text to display in the task interface (defaults to `Getting Hashicorp Vaults`)

    ??? success "Output"

        A list of [Vaults](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.hashicorp_vaults:
          save: vaults
        ```"""
    return _get_by_driver(reason, Drivers.hashicorp.value)


def hashicorp_vault_by_name(name: str) -> Vault:
    """Get the Hashicorp Vault with the supplied name.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __name:__ the name of the Hashicorp Vault

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.hashicorp_vault_by_name:
            name: PPA Vault
          save: vault
        ```"""
    return _get_by_name(name, vault_type=Vaults.hashicorp.value, getter=hashicorp_vaults)


def pam_vaults(reason: str = f"Getting {Vaults.pam.value}s") -> List[Vault]:
    """Get all Osirium PAM Vaults configured in PPA.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __reason:__ text to display in the task interface (defaults to `Getting Osirium PAM Vaults`)

    ??? success "Output"

        A list of [Vaults](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.pam_vaults:
          save: vaults
        ```"""
    return _get_by_driver(reason, Drivers.pam.value)


def pam_vault_by_name(name: str) -> Vault:
    """Get the Osirium PAM Vault with the supplied name.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __name:__ the name of the Osirium PAM Vault

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.pam_vault_by_name:
            name: PAM Appliance
          save: vault
        ```"""
    return _get_by_name(name, vault_type=Vaults.pam.value, getter=pam_vaults)


@min_ppa_version("2.7.1")
def aws_vaults(reason: str = f"Getting {Vaults.aws.value}s") -> List[Vault]:
    """Get all AWS Secrets Manager Vaults configured in PPA.

    !!! question "Minimum Plugin Version: 5.5.0"

    !!! question "Minimum PPA Version: 2.7.1"

    ??? info "Input"

        - __reason:__ text to display in the task interface (defaults to `Getting AWS Secrets Manager Vaults`)

    ??? success "Output"

        A list of [Vaults](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.aws_vaults:
          save: vaults
        ```"""
    return _get_by_driver(reason, Drivers.aws.value)


@min_ppa_version("2.7.1")
def aws_vault_by_name(name: str) -> Vault:
    """Get the AWS Secrets Manager Vault with the supplied name.

    !!! question "Minimum Plugin Version: 5.5.0"

    !!! question "Minimum PPA Version: 2.7.1"

    ??? info "Input"

        - __name:__ the name of the AWS Secrets Manager Vault

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.aws_vault_by_name:
            name: AWS EU West 2
          save: vault
        ```"""
    return _get_by_name(name, vault_type=Vaults.aws.value, getter=aws_vaults)


@min_ppa_version("2.7.0")
def azure_vaults(reason: str = f"Getting {Vaults.azure.value}s") -> List[Vault]:
    """Get all Azure Key Vaults configured in PPA.

    !!! question "Minimum Plugin Version: 5.5.0"

    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __reason:__ text to display in the task interface (defaults to `Getting Azure Key Vaults`)

    ??? success "Output"

        A list of [Vaults](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.azure_vaults:
          save: vaults
        ```"""
    return _get_by_driver(reason, Drivers.azure.value)


@min_ppa_version("2.7.0")
def azure_vault_by_name(name: str) -> Vault:
    """Get the Azure Key Vault with the supplied name.

    !!! question "Minimum Plugin Version: 5.5.0"

    !!! question "Minimum PPA Version: 2.7.0"

    ??? info "Input"

        - __name:__ the name of the Azure Key Vault

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.azure_vault_by_name:
            name: Azure
          save: vault
        ```"""
    return _get_by_name(name, vault_type=Vaults.azure.value, getter=azure_vaults)


def output_table(text: str, vaults: Union[Vault, List[Vault]]) -> None:
    """Display vaults in a table.

    The table will have the following columns:

    - Name
    - Type
    - Address
    - Expires At

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __vaults:__ a single or list of [Vaults](../inputs_outputs/#vault) to display in the table

    ??? success "Output"

        Nothing is outputted by this action.

    ??? example
        ```YAML linenums="1"
        - ppa.vaults.get_all:
          save: vaults

        - ppa.vaults.output_table:
            text: All Vaults
          load:
            vaults: vaults
        ```
    """
    vaults = [vaults] if not isinstance(vaults, list) else vaults
    ui.output_table(text, header=TABLE_HEADER, data=[_generate_table_row(v) for v in vaults])


def input_table(
    text: str, vaults: Union[Vault, List[Vault]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> List[Vault]:
    """Display vaults in a table, & allow the task operator to make a selection.

    The table will have the following columns:

    - Name
    - Type
    - Address
    - Expires At

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __text:__ the title of the table

        - __vaults:__ a single or list of [Vaults](../inputs_outputs/#vault) to display in the table

        - __minimum:__ the minimum number of acceptable selections

        - __maximum:__ the maximum number of acceptable selections

    ??? success "Output"

        A list of [Vaults](../inputs_outputs/#vault).

    ??? tip

        - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
        - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
            of selections.

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.hashicorp_vaults:
          save: vaults

        - ppa.vaults.input_table:
            text: Select a Hashicorp Vault
            minimum: 1
            maximum: 1
          load:
            vaults: vaults
          save: vault
        ```
    """
    vaults = [vaults] if not isinstance(vaults, list) else vaults
    table_id, error_id = list(map(lambda _: str(uuid.uuid4()), range(0, 2)))
    while True:
        indices = ui.input_table(text, header=TABLE_HEADER, data=[_generate_table_row(v) for v in vaults], element_id=table_id,)
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} Vault.", element_id=error_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} Vault(s).", element_id=error_id)
            continue
        break
    return [vaults[i] for i in indices]


def _select_vault(vault_type: Vaults, *, getter: Callable, auto_select) -> Vault:
    if vaults := getter():
        if all([len(vaults) == 1, auto_select]):
            return next(iter(vaults))
        return next(iter(input_table(f"Select {vault_type.value}", vaults, minimum=1, maximum=1)))
    _raise(f"There are no {vault_type.value}s configured in PPA.", exception=KeyError)


def select_pam_vault(auto_select: Optional[bool] = True) -> Vault:
    """Select a PAM vault from the PPA configuration.

    By default if only 1 PAM vault exists, it will be selected automatically.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __auto_select:__ set to _false_ to force table selection if only 1 vault exists (defaults to _true_)

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.select_pam_vault:
          save: vault
        ```
    """
    return _select_vault(Vaults.pam, getter=pam_vaults, auto_select=auto_select)


def select_hashicorp_vault(auto_select: Optional[bool] = True) -> Vault:
    """Select a Hashicorp vault from the PPA configuration.

    By default if only 1 Hashicorp vault exists, it will be selected automatically.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __auto_select:__ set to _false_ to force table selection if only 1 vault exists (defaults to _true_)

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.select_hashicorp_vault:
          save: vault
        ```
    """
    return _select_vault(Vaults.hashicorp, getter=hashicorp_vaults, auto_select=auto_select)


@min_ppa_version("2.7.1")
def select_aws_vault(auto_select: Optional[bool] = True) -> Vault:
    """Select an AWS Secrets Manager vault from the PPA configuration.

    By default if only 1 AWS Secrets Manager vault exists, it will be selected automatically.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __auto_select:__ set to _false_ to force table selection if only 1 vault exists (defaults to _true_)

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.select_aws_vault:
          save: vault
        ```
    """
    return _select_vault(Vaults.aws, getter=aws_vaults, auto_select=auto_select)


@min_ppa_version("2.7.0")
def select_azure_vault(auto_select: Optional[bool] = True) -> Vault:
    """Select an Azure Key vault from the PPA configuration.

    By default if only 1 Azure Key vault exists, it will be selected automatically.

    !!! question "Minimum Plugin Version: 5.5.0"

    ??? info "Input"

        - __auto_select:__ set to _false_ to force table selection if only 1 vault exists (defaults to _true_)

    ??? success "Output"

        A single [Vault](../inputs_outputs/#vault).

    ??? example

        ```YAML linenums="1"
        - ppa.vaults.select_azure_vault:
          save: vault
        ```
    """
    return _select_vault(Vaults.azure, getter=azure_vaults, auto_select=auto_select)
